/*
 * Kalisto
 *
 * Copyright (c) 2001-2005
 * 	Distributed Systems Research Group
 * 	Department of Software Engineering
 * 	Faculty of Mathematics and Physics
 * 	Charles University, Czech Republic
 * see ../README for more details
 *
 *
 * System start-up
 *
 * This module only has a single routine located at (hardwired) virtual
 * address 0xbfc00000, which is where the processor jumps after reset. The
 * address is located in a part of address space which is not cached and not
 * mapped through TLB, instead it is mapped directly (hardwired) to translate
 * to physical addresses obtained by substracting 0x8000000 from the virtual
 * address.
 *
 * For this particular address, the bootstrap code has to be located at
 * address 0x1fc00000 in physical memory. The reason for this is to allow
 * the bootstrap code to run without need to worry about uninitialized TLB
 * and to possibly check the cache operation, before passing control to
 * the kernel code placed in a cached part of the address space.
 *
 * On entry, the contents of all general registers is undefined. The BEV and
 * EXL bits in the Status Register are set to 1, which means that the
 * processor operates in kernel mode with (implicitly) disabled interrupts.
 */

	.text

	.set noat
	.set noreorder
	.set nomacro

#include "regname.h"

/*
 * The kernel startup address consist of the physical address 0x00000300 and
 * the operating mode flags  - the three highest bits - KSEG0 (unmapped,
 * cached).
 */
#define KERNEL_STARTUP_ADDRESS	0x80000300

	.ent	__start
__start:

	/*
	 * Jump to the kernel image entry point. Note that there is no
	 * instruction in the MIPS R4000 instruction set to load a 32bit
	 * constant. You can use the "la" macro instruction to make the
	 * assembler issue the two required instructions for you.
	 */
	lui	$ra, KERNEL_STARTUP_ADDRESS >> 16
	ori	$ra, KERNEL_STARTUP_ADDRESS & 0xffff

	j	$ra
	nop

	.end __start

